/******************************************************************************
/*   This file is part of the game Various Burst.
 *
 *   Various Burst is free software: you can redistribute it and/or modify
 *   it under the terms of the WTF Public License.
 *
 *   Various Burst is distributed in the hope that players will find it
 *   enjoyable, emotive, educational even, but with NO WARRANTY of any
 *   kind.  What happens to your computer, your dog or your sanity as
 *   a result of this is solely YOUR responsibility.
 *
 *   You should have received a copy of the WTF Public License
 *   along with the game (please see the file COPYING).  If not,
 *   please refer to <http://www.wtfpl.net/about/>.
 */

#include "common.h"
#include "blenders.h"
#include "backdrop.h"

/*! @defgroup backdrop_module "Backdrops"
 *  @brief This module contains functionality to draw the scrolling backgrounds.
 *  @{
 */
#define BACKDROP_MAX_BACKDROPS   4
#define BACKDROP_FADE_START      30
#define BACKDROP_NUM_STARS       32

/*! @brief a structure for handling the scrolling stars in the background.
 */
typedef struct
{
   int            x;
   int            y;
   unsigned char  colour;
} BACKDROP_STAR;

/*! @brief the first starfield scroll plane, scrolled at 1/2 speed; private. */
static BACKDROP_STAR backdrop_starfield1[BACKDROP_NUM_STARS];
/*! @brief the first starfield scroll plane, scrolled at 1/3 speed; private. */
static BACKDROP_STAR backdrop_starfield2[BACKDROP_NUM_STARS];

/*! @brief checks whether we've randomised the stars prior to painting them
 * the first time; should be considered module-private.
 */
static BOOL backdrop_stars_inited      = FALSE;

/*! @brief the currently-displayed backdrop image. should be considered private.
 */
static BITMAP *backdrop_curr = NULL;

/*! @brief the previously-displayed backdrop; kept in memory to make it easy
 *   to crossfade from it to the 'current' one. should be considered
 *   module-private.
 */
static BITMAP *backdrop_prev = NULL;

/*! @brief animation clock to help keep track of the fading animation - should
 * be considered module private.
 */
static unsigned char backdrop_fade_value = 0;

/*! @brief animation clock to keep track of the scrolling animation - should be
 * considered module private.
 */
static unsigned int  backdrop_scroll_value = 0;

/******************************************************************************/
/*! @brief Load the next backdrop image that should appear into memory and ready
 *  it for displaying.
 *  @note Also signals the rest of the module to tell it to begin crossfading
 *  from the previous image to the current one.
 *  @param which Which backdrop to load, automagically clamped to a safe range.
 */
void BACKDROP_load_new(unsigned which)
{
   char buf[FILENAME_BUFFER_LENGTH];

   snprintf(buf, FILENAME_BUFFER_LENGTH, "./gfx/backdrop/%02x.pcx",
      which % BACKDROP_MAX_BACKDROPS);

   if (backdrop_prev != NULL) destroy_bitmap(backdrop_prev);

   backdrop_prev = backdrop_curr;

   backdrop_fade_value = BACKDROP_FADE_START;

   if (exists(buf))
   {
      backdrop_curr = load_bitmap(buf, NULL);
   }
   else
   {
      backdrop_curr = create_bitmap(DISPLAY_WIDTH, DISPLAY_HEIGHT);
      clear_bitmap(backdrop_curr);
      textout_ex(backdrop_curr, font, "missing image", 16, 16, -1, -1);
   }
}

/******************************************************************************/
/*! @brief Animate, process and draw the star scrolling planes; module-private.
 */
static void BACKDROP_draw_stars(void)
{
   int index;
   unsigned should_scroll_plane_1 = (backdrop_scroll_value % 2);
   unsigned should_scroll_plane_2 = (backdrop_scroll_value % 3);

   // make sure these get randomised a bit before drawing them the first time,
   // otherwise, they'll be stuck on the same y value...
   if (!backdrop_stars_inited)
   {
      for (index = 0; index < BACKDROP_NUM_STARS; index++)
      {
         backdrop_starfield1[index].y        = rand() % DISPLAY_HEIGHT;
         backdrop_starfield1[index].x        = rand() % DISPLAY_WIDTH;
         backdrop_starfield1[index].colour   = rand();

         backdrop_starfield2[index].y        = rand() % DISPLAY_HEIGHT;
         backdrop_starfield2[index].x        = rand() % DISPLAY_WIDTH;
         backdrop_starfield2[index].colour   = rand();

         backdrop_stars_inited = TRUE;
      }
   }

   // process animation and resetting
   if (should_scroll_plane_1)
   {
      for (index = 0; index < BACKDROP_NUM_STARS; index++)
      {
         backdrop_starfield1[index].y++;

         if (backdrop_starfield1[index].y > DISPLAY_HEIGHT)
         {
            backdrop_starfield1[index].y        = 0;
            backdrop_starfield1[index].x        = rand() % DISPLAY_WIDTH;
            backdrop_starfield1[index].colour   = rand();
         }
      }
   }

   if (should_scroll_plane_2)
   {
      for (index = 0; index < BACKDROP_NUM_STARS; index++)
      {
         backdrop_starfield2[index].y++;

         if (backdrop_starfield2[index].y > DISPLAY_HEIGHT)
         {
            backdrop_starfield2[index].y        = 0;
            backdrop_starfield2[index].x        = rand() % DISPLAY_WIDTH;
            backdrop_starfield2[index].colour   = rand();
         }
      }
   }

   // paint them all
   drawing_mode(DRAW_MODE_TRANS, NULL, 0, 0);
   set_blender_mode(NULL, (BLENDER_FUNC)BLEND_add16, NULL, 32, 32, 32, 32);

   for (index = 0; index < BACKDROP_NUM_STARS; index++)
   {
      putpixel(double_buffer,
         backdrop_starfield1[index].x, backdrop_starfield1[index].y,
         makecol(backdrop_starfield1[index].colour, backdrop_starfield1[index].colour, backdrop_starfield1[index].colour));
   }

   set_blender_mode(NULL, (BLENDER_FUNC)BLEND_add16, NULL, 16, 16, 16, 16);

   for (index = 0; index < BACKDROP_NUM_STARS; index++)
   {
      putpixel(double_buffer,
         backdrop_starfield2[index].x, backdrop_starfield2[index].y,
         makecol(backdrop_starfield2[index].colour, backdrop_starfield2[index].colour, backdrop_starfield2[index].colour));
   }

   solid_mode();
}

/******************************************************************************/
/*! @brief Repaint the background. Also has the side effect of advancing the
 *  fade and scroll animations.
 */
void BACKDROP_draw(void)
{
   backdrop_scroll_value++;

   // are we still crossfading?
   if (backdrop_fade_value > 0)
   {
      // <konata inflection="haruhi"> Orders from the Chief. Wait here! </konata>
      backdrop_fade_value--;

      if (backdrop_prev != NULL)
      {
         blit(backdrop_prev, double_buffer, 0, 0,
            0, backdrop_scroll_value % backdrop_prev->h,
            backdrop_prev->w, backdrop_prev->h);

         blit(backdrop_prev, double_buffer, 0, 0,
            0, (backdrop_scroll_value % backdrop_prev->h) - backdrop_prev->h,
            backdrop_prev->w, backdrop_prev->h);

         if (backdrop_prev->h < DISPLAY_HEIGHT)
         {
            blit(backdrop_prev, double_buffer, 0, 0,
               0, (backdrop_scroll_value % backdrop_prev->h) + backdrop_prev->h,
               backdrop_prev->w, backdrop_prev->h);
         }
      }

      if (backdrop_curr != NULL)
      {
         set_trans_blender(255, 255, 255,
            (int)(255.0f * (BACKDROP_FADE_START - backdrop_fade_value) / (float)BACKDROP_FADE_START));

         draw_trans_sprite(double_buffer, backdrop_curr, 0,
            backdrop_scroll_value % backdrop_curr->h);
         draw_trans_sprite(double_buffer, backdrop_curr, 0,
            (backdrop_scroll_value % backdrop_curr->h) - backdrop_curr->h);

         if (backdrop_curr->h < DISPLAY_HEIGHT)
         {
            draw_trans_sprite(double_buffer, backdrop_curr, 0,
               (backdrop_scroll_value % backdrop_curr->h) + backdrop_curr->h);
         }
      }
   }
   else
   {
      // no.
      if (backdrop_curr != NULL)
      {
         blit(backdrop_curr, double_buffer, 0, 0,
            0, backdrop_scroll_value % backdrop_curr->h,
            backdrop_curr->w, backdrop_curr->h);

         blit(backdrop_curr, double_buffer, 0, 0,
            0, (backdrop_scroll_value % backdrop_curr->h) - backdrop_curr->h,
            backdrop_curr->w, backdrop_curr->h);

         if (backdrop_curr->h < DISPLAY_HEIGHT)
         {
            blit(backdrop_curr, double_buffer, 0, 0,
               0, (backdrop_scroll_value % backdrop_curr->h) + backdrop_curr->h,
               backdrop_curr->w, backdrop_curr->h);
         }
      }
   }

   BACKDROP_draw_stars();
}


/******************************************************************************/
/*! @brief Free up any loaded images.
 */
void BACKDROP_unload(void)
{
   if (backdrop_prev != NULL) destroy_bitmap(backdrop_prev);
   if (backdrop_curr != NULL) destroy_bitmap(backdrop_curr);

   backdrop_prev = NULL;
   backdrop_curr = NULL;
}

/*! @} */
