/******************************************************************************
/*   This file is part of the game Various Burst.
 *
 *   Various Burst is free software: you can redistribute it and/or modify it
 *   under the terms of the WTF Public License.
 *
 *   Various Burst is distributed in the hope that players will find it enjoyable,
 *   emotive, educational even, but with NO WARRANTY of any kind. What happens
 *   to your Pandora, your dog or your sanity as a result of this is solely
 *   YOUR responsibility.
 *
 *   You should have received a copy of the WTF Public License along with the
 *   source code to the game (please see the file COPYING).  If not, please refer to
 *   <http://www.wtfpl.net/about/>.
 */

#include "common.h"
#include "blenders.h"
/*!
 * \defgroup blenders_c Custom blending functions
 * \brief These are blending functions that Allegro uses as a
 * callback to do colour arithmetic and create some of the visual effects in the
 * game.
 * @{
 */

/******************************************************************************/
/*!
 * @brief Multiply a background pixel by (factor/64)
 *   to darken it.
 *
 * @param src Incoming background pixel colour
 * @param dst Incoming background pixel colour
 * @param factor How much light this pixel should
 *   have applied to it.
 */
inline uint16_t BLEND_lightmap16(uint16_t dst, uint16_t src, uint16_t factor)
{
   // workaround for an inconsistency in Allegro -
   // draw_lit_sprite() fills dst, but draw_gouraud_sprite()
   // fills src. Since we'll be completely ignoring the
   // destination here, we'll simply discard one of these,
   // since it'll be meaningless anyway
   if(src != 0) dst = src;

   // trivial cases: max brightness (just return the start colour)
   // min brightness (just return 0);
   if(factor >= 64) return dst;
   if(factor < 1)   return 0;

   // split dst into its components and multiply by lightmap
   // value...
   uint16_t dst_rb = (((dst & 0xF81f) * factor) >> 6) & 0xF81F;
   uint16_t dst_g  = (((dst & 0x07E0) * factor) >> 6) & 0x07E0;
   return (dst_rb | dst_g);
}

/******************************************************************************/
/*!
 * @brief Desaturate a background pixel by an amount
 * controlled by the brightness of the green channel
 * of a foreground pixel.
 *
 * @note the factor aregument is ignored!
 *
 * @param src Incoming background pixel colour
 * @param dst Incoming foreground pixel colour
 * @param factor Ignored.
 */
inline uint16_t BLEND_desat16(uint16_t dst, uint16_t src, uint16_t factor)
{
   uint16_t r, g, b;
   uint16_t rgb_quotient;

   // Trivial cases: desaturating a black or white
   // pixel has no effect
   if ((src == 0) || (src == 0xFFFF)) return src;

   // incoming colour from foreground - interpreted as
   // blend intensity. We use the green component since
   // it has the most precision
   unsigned long blend_fact = (dst & 0x07E0) >> 5;

   // trivial case: blend intensity factor is 0, so
   // we don't need to do anything to this pixel
   if (blend_fact == 0) return src;

   // incoming colour from background
   r = (src & 0xF800) >> 11;
   g = (src & 0x07E0) >> 5;
   b = (src & 0x1F);

   // convert background colour to greyscale (doesn't take into
   // account luminosity, kinda inaccurate, but meh)
   rgb_quotient = (r + g + b) >> 1;

   // outgoing pixel:
   //   (src * (max value - blend factor)) + (dest * blend factor))
   //   -----------------------------------------------------------
   //                          max value

   // trivial case: blend intensity factor is
   // all the way - this pixel is fully greyscale
   if (blend_fact == 0x3F)
   {
      return ((rgb_quotient << 10) & 0xF800) |
         ((rgb_quotient << 5) & 0x07E0)      |
         ((rgb_quotient >> 1) & 0x001F);
   }

   unsigned long ret_r;
   unsigned long ret_g;
   unsigned long ret_b;

   ret_r = ((r * (64 - blend_fact)) + ((rgb_quotient >> 1) * blend_fact)) >> 6;
   ret_g = ((g * (64 - blend_fact)) + (rgb_quotient * blend_fact)) >> 6;
   ret_b = ((b * (64 - blend_fact)) + ((rgb_quotient >> 1) * blend_fact)) >> 6;

   if (ret_r > 63) ret_r = 63;
   if (ret_g > 63) ret_g = 63;
   if (ret_b > 63) ret_b = 63;

   return ((ret_r << 11) & 0xF800) |
      ((ret_g << 5) & 0x07E0)      |
      (ret_b & 0x001F);
}

/******************************************************************************/
/*!
 * @brief Add (dest colour * (factor/max intensity)) + (src colour)
 * and return this sum, clamped between 0 and maximum intensity
 *
 * @param src Incoming background pixel colour
 * @param dst Incoming foreground pixel colour
 * @param factor Controls what fraction of dst should be added
 * to src. Should be a number between 0 and 64 inclusive.
 */
inline uint16_t BLEND_add16(uint16_t dst, uint16_t src, uint16_t factor)
{
   // Clamp factor to allowed range.
   // Trivial case: adding with a blend intensity
   // of 0 has no effect.
   if (factor < 1) return src;
   if (factor > 64) factor = 64;

   // Incoming source pixel.
   uint16_t src_rb = src & 0xF81F;
   uint16_t src_g = src & 0x07E0;

   // Trivial case: adding anything to a white pixel
   // has no effect.
   if (src == 0xFFFF) return src;

   // Incoming dest pixel, multiplied by factor.
   unsigned long dst_rb = (((dst & 0xF81F) * factor) >> 6) & 0xF81F;
   unsigned long dst_g = (((dst & 0x07E0) * factor) >> 6) & 0x07E0;

   // Compute sums and clamp to allowed ranges.
   unsigned long ret_rb = src_rb + dst_rb;
   unsigned long ret_g = src_g + dst_g;

   // Saturate red.
   if(ret_rb & 0xFFFF0000) ret_rb = ((ret_rb | 0xF800));

   // Saturate blue.
   if(ret_rb & 0x07E0) ret_rb = ((ret_rb | 0x001F));

   // Saturate and clamp green.
   if(ret_g & 0xFFFFF800) ret_g = ((ret_g | 0x07E0) & 0x07E0);

   // Clamp blue & red.
   ret_rb = ret_rb & 0xF81F;

   return (ret_rb | ret_g);
}

/******************************************************************************/
/*!
 * @brief Subtract  (src colour) - (dest colour * (factor/max intensity))
 * and return this difference, clamped between 0 and maximum intensity
 *
 * @param src Incoming background pixel colour
 * @param dst Incoming foreground pixel colour
 * @param factor Controls what fraction of dst should be subtracted
 * from src. Should be a number between 0 and 64 inclusive.
 */
inline uint16_t BLEND_sub16(uint16_t dst, uint16_t src, uint16_t factor)
{
   // Clamp factor to allowed range.
   // Trivial case: subtracting with a blend intensity
   // of 0 has no effect.
   if (factor < 1) return src;
   if (factor > 64) factor = 64;

   // Incoming source pixel.
   uint16_t src_r = src & 0xF800;
   uint16_t src_g = src & 0x07E0;
   uint16_t src_b = src & 0x001F;

   // Trivial case: subtracting anything from a black
   // pixel has no effect.
   if (src == 0x0) return src;

   // Incoming dest pixel, multiplied by factor.
   uint16_t dst_r = (((dst & 0xF800) * factor) >> 6) & 0xF800;
   uint16_t dst_g = (((dst & 0x07E0) * factor) >> 6) & 0x07E0;
   uint16_t dst_b = (((dst & 0x001F) * factor) >> 6) & 0x001F;

   // Compute sums and clamp to allowed ranges.
   if (src_r < dst_r) dst_r = src_r;
   if (src_g < dst_g) dst_g = src_g;
   if (src_b < dst_b) dst_b = src_b;

   uint16_t ret_r = (src_r - dst_r);
   uint16_t ret_g = (src_g - dst_g);
   uint16_t ret_b = (src_b - dst_b);

   return (ret_r | ret_g | ret_b);
}

/*! @} */
