/******************************************************************************
/*   This file is part of the game Various Burst.
 *
 *   Various Burst is free software: you can redistribute it and/or modify it
 *   under the terms of the WTF Public License.
 *
 *   Various Burst is distributed in the hope that players will find it enjoyable,
 *   emotive, educational even, but with NO WARRANTY of any kind. What happens
 *   to your Pandora, your dog or your sanity as a result of this is solely
 *   YOUR responsibility.
 *
 *   You should have received a copy of the WTF Public License along with the
 *   source code to the game (please see the file COPYING).  If not, please refer to
 *   <http://www.wtfpl.net/about/>.
 */
#include "blurrer.h"

// private data
static BITMAP *blur_buffer = NULL;

// private funcs
void BLUR_init(void);
void BLUR_cleanup(void);

/*******************************************************************/
/*!
 * @brief creates the blurrer's buffer, and ensures it'll
 * be automatically deleted at the end of runtime.
 */
void BLUR_init(void)
{
   if (blur_buffer == NULL)
   {
      blur_buffer = create_bitmap(DISPLAY_WIDTH, DISPLAY_HEIGHT);
      clear_bitmap(blur_buffer);
      atexit(BLUR_cleanup);
   }
}

/*******************************************************************/
/*!
 * @brief frees the blurrer's buffer; should not need to be called
 * manually.
 */
void BLUR_cleanup(void)
{
   if (blur_buffer != NULL)
   {
      destroy_bitmap(blur_buffer);
      blur_buffer = NULL;
   }
}

/*******************************************************************/
/*!
 * @brief   Applies the blur effect to the destination bitmap.
 * @warning Due to developer laziness, the destination bitmap MUST
 *  contain DISPLAY_WIDTH * DISPLAY_HEIGHT pixels, or the app WILL
 *  crash.
 */
void BLUR_apply(BITMAP *dest)
{
   if (blur_buffer == NULL)
   {
      BLUR_init();
   }

   int x_index;
   int y_index;

   unsigned short up_neighbour;
   unsigned short dn_neighbour;
   unsigned short lt_neighbour;
   unsigned short rt_neighbour;
   unsigned short centre;

   unsigned int   rb_tmp;
   unsigned int   g_tmp;

   // explanation of the crazy bitmasks: because the results of the red
   // and blue addition and division are separated by the green field, we
   // can process them together, reducing the amount of arithmetic that has
   // to be done each pixel (0xf81f = 11111 000000 11111).

   // someone smarter than me who knows ARM assembler could probably speed
   // this up...

   for (y_index = 1; y_index < (DISPLAY_HEIGHT - 1); y_index++)
   {
      for (x_index = 1; x_index < (DISPLAY_WIDTH - 1); x_index++)
      {
         // sample the centre and the four neighbouring pixels
         up_neighbour = ((unsigned short *)dest->line[y_index - 1])[x_index + 0];
         dn_neighbour = ((unsigned short *)dest->line[y_index + 1])[x_index + 0];
         lt_neighbour = ((unsigned short *)dest->line[y_index + 0])[x_index - 1];
         rt_neighbour = ((unsigned short *)dest->line[y_index + 0])[x_index + 1];

         // sum the neighbour colour components, divide by four
         rb_tmp = (up_neighbour & 0xf81f)
                  + (dn_neighbour & 0xf81f)
                  + (lt_neighbour & 0xf81f)
                  + (rt_neighbour & 0xf81f);

         g_tmp = (up_neighbour & 0x07e0)
                  + (dn_neighbour & 0x07e0)
                  + (lt_neighbour & 0x07e0)
                  + (rt_neighbour & 0x07e0);

         rb_tmp  = (rb_tmp >> 2) & 0xf81f;
         g_tmp   = (g_tmp  >> 2) & 0x07e0;

         ((unsigned short *)blur_buffer->line[y_index])[x_index] = rb_tmp | g_tmp;
      }
   }

   blit(blur_buffer, dest, 0,0,0,0, DISPLAY_WIDTH, DISPLAY_HEIGHT);
}
