/******************************************************************************
/*   This file is part of the game Various Burst.
 *
 *   Various Burst is free software: you can redistribute it and/or modify
 *   it under the terms of the WTF Public License.
 *
 *   Various Burst is distributed in the hope that players will find it
 *   enjoyable, emotive, educational even, but with NO WARRANTY of any
 *   kind.  What happens to your computer, your dog or your sanity as
 *   a result of this is solely YOUR responsibility.
 *
 *   You should have received a copy of the WTF Public License
 *   along with the game (please see the file COPYING).  If not,
 *   please refer to <http://www.wtfpl.net/about/>.
 */

#include "common.h"

/*****************************************************************************/

BITMAP          *double_buffer  = NULL;

DUH             *COMMON_duh     = NULL;
AL_DUH_PLAYER   *COMMON_duhplyr = NULL;
float           COMMON_duhvol   = DUMB_VOL;
int             COMMON_should_fade_bgm;
static int      COMMON_curr_bgm = 0;

int             bgm_fade_done   = FALSE;
int             next_state;
volatile int    frames_this_sec;
volatile int    frames_per_sec;
volatile int    time_for_logic;
BOOL           curr_state_done;

#ifdef   __OPENPANDORA__
int             fbdev;
int             source_addr;
void            *scrn_ptr;
#endif


/******************************************************************************/

void sample_framerate(void)
{
   frames_per_sec  = frames_this_sec;
   frames_this_sec = 0;
}
END_OF_FUNCTION(sample_framerate);

/******************************************************************************/

void game_heartbeat()
{
   time_for_logic = 1;
}
END_OF_FUNCTION(game_heartbeat);

/******************************************************************************/
/*! @brief Create the window, set the video mode, allocate the default font and
 *  do any other required initialization.
 */

void COMMON_setup(void)
{
   // Set up the usual stuff...
   allegro_init();
   install_keyboard();
   install_joystick(JOY_TYPE_AUTODETECT);
   install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL);
   atexit(&dumb_exit);
   dumb_register_stdfiles();
   dumb_resampling_quality = DUMB_RQ_LINEAR;

   // Set up graphics modes.
   set_color_depth(16);

   int tmp_w;
   int tmp_h;

   get_desktop_resolution(&tmp_w, &tmp_h);

   float desktop_aspect_ratio = (float)tmp_h / (float)tmp_w;

/*
   #ifndef ARCADE_MACHINE_BUILD
   {
      set_gfx_mode(GFX_AUTODETECT_WINDOWED, 640, 480, 0, 0);
      set_display_switch_mode(SWITCH_BACKGROUND);
   }
   #else
   {
      set_gfx_mode(GFX_AUTODETECT_FULLSCREEN, 640, 480, 0, 0);
      set_display_switch_mode(SWITCH_NONE);
   }
   #endif
*/

#ifdef __OPENPANDORA__
    set_gfx_mode(GFX_AUTODETECT, DISPLAY_WIDTH, DISPLAY_HEIGHT, 0, 0);

    system("ofbset -fb /dev/fb1 -pos 0 0 -size 800 480 -mem 192000 -en 1 ");
    system("fbset -fb /dev/fb1 -g 400 240 400 240 16");

    fbdev = open("/dev/fb1", O_RDWR);

    if(fbdev == -1) exit(2);

    scrn_ptr = mmap(0,400 * 240 * 2, PROT_WRITE | PROT_READ, MAP_SHARED, fbdev, 0);

    if(scrn_ptr == MAP_FAILED)
    {
        allegro_message("PANIC: Couldn't set graphics mode!");
        exit(1);
    }

#else
    set_gfx_mode(GFX_AUTODETECT_WINDOWED, DISPLAY_WIDTH * 2, DISPLAY_HEIGHT * 2, 0, 0);
#endif

#ifdef __WIN32__
    // fix for problem with game freezing when its
    // window does not have focus.
    set_display_switch_mode(SWITCH_BACKGROUND);
#endif

   // We don't check for an error when creating this, because
   // if it fails, chances are we're out of memory and can't
   // really die gracefully or play the game, and the OOM killer
   // is probably gonna get triggered anyway...
   double_buffer = create_bitmap(DISPLAY_WIDTH, DISPLAY_HEIGHT);

#ifdef __OPENPANDORA__
    source_addr = bmp_read_line(double_buffer,0);
#endif

   // Replace the default font with our own, if possible.
   if (exists(DEFAULT_FONT_PATH))
   {
      font = load_font(DEFAULT_FONT_PATH,NULL,NULL);
   }

   // Game's main timer
   time_for_logic = 0;
   LOCK_VARIABLE(time_for_logic);
   LOCK_FUNCTION(game_heartbeat);

   install_int_ex(game_heartbeat,BPS_TO_TIMER(TICKS_PER_SEC));

   // Performance sampling
   LOCK_VARIABLE(frames_per_sec);
   LOCK_VARIABLE(frames_this_sec);
   LOCK_FUNCTION(sample_framerate);

   install_int_ex(sample_framerate,BPS_TO_TIMER(1));
}

/******************************************************************************/
/*! Paint whatever's currently in the double buffer to the screen. We also
 *  poll the music player and handle volume fades here.
 */
void COMMON_flip_buffer(void)
{
   if(COMMON_duhplyr != NULL)
   {
      // check if a music change is pending
      if(COMMON_should_fade_bgm)
      {
         // yes, fade out
         COMMON_duhvol = COMMON_duhvol * DUMB_FADE_MUL;
         al_duh_set_volume(COMMON_duhplyr, COMMON_duhvol);

         if(COMMON_duhvol < DUMB_FADE_MIN)
         {
            bgm_fade_done = TRUE;
         }
      }

      al_poll_duh(COMMON_duhplyr);
   }

#ifdef __OPENPANDORA__
    memmove(scrn_ptr, (int *)source_addr, DISPLAY_WIDTH * DISPLAY_HEIGHT * 2);
#else
    stretch_blit(double_buffer,screen,0,0,double_buffer->w,double_buffer->h,
                 0,0, screen->w, screen->h);
#endif
   frames_this_sec++;
}

/******************************************************************************/
/*! @brief Display a performance counter that shows how many frames were
 *   drawn in the last second.
 *  @note Should not be used in release builds...
 */
void COMMON_show_fps(void)
{
   textprintf_ex(double_buffer, font, 8, 8, 0xFFFF, 0, "FPS: %02d", frames_per_sec);
}

/*****************************************************************************/
/*! Clean up after the program.
 */
void COMMON_shutdown(void)
{
    clear_keybuf();
    remove_int(game_heartbeat);
    remove_int(sample_framerate);
    if(double_buffer != NULL) destroy_bitmap(double_buffer);
    if(COMMON_duh != NULL) unload_duh(COMMON_duh);
    set_gfx_mode(GFX_TEXT,0,0,0,0);

#ifdef __OPENPANDORA__
    munmap(scrn_ptr, 400*240*2);
    close(fbdev);
    system("ofbset -fb /dev/fb1 -pos 0 0 -size 0 0 -mem 0 -en 0");

#endif
}

/*****************************************************************************/
/*! @brief Change what music is playing. To stop BGM entirely, pass a non-
 * existant track ID.
 */
void COMMON_set_bgm(int track)
{
   char buffer[2048];

   // already playing this, don't need to do anything
   if(track == COMMON_curr_bgm) return;

   COMMON_curr_bgm = track;

   // check whether requested music exists
   sprintf(buffer,"./bgm/%02d.xm",track);
   if(exists(buffer))
   {
      if (COMMON_duhplyr != NULL)
         al_stop_duh(COMMON_duhplyr);

      // yes, load it and start playing
      COMMON_duh              = dumb_load_xm_quick(buffer);
      COMMON_duhplyr          = al_start_duh(COMMON_duh, 2, 0, DUMB_VOL, DUMB_BUFSIZE, DUMB_MIXRATE);
      bgm_fade_done           = FALSE;
      COMMON_duhvol           = DUMB_VOL;
      COMMON_should_fade_bgm  = FALSE;
   }
   else
   {
      // no - null out duh and player
      if (COMMON_duhplyr != NULL)
         al_stop_duh(COMMON_duhplyr);
      COMMON_duhplyr = NULL;
   }
}

/*****************************************************************************/
/*! @brief Cause the currently-playing track to fade out.
 *  Returns immediately and does not block; the actual act of fading happens
 *  in COMMON_flip_buffer.
 */
void COMMON_fade_bgm(void)
{
   COMMON_should_fade_bgm = TRUE;
}

/*****************************************************************************/
/*! @brief Get the track ID of the currently-playing BGM.
 */
int COMMON_get_bgm(void)
{
   return COMMON_curr_bgm;
}
