/******************************************************************************
/*   This file is part of the game Various Burst.
 *
 *   Various Burst is free software: you can redistribute it and/or modify it
 *   under the terms of the WTF Public License.
 *
 *   Various Burst is distributed in the hope that players will find it enjoyable,
 *   emotive, educational even, but with NO WARRANTY of any kind. What happens
 *   to your Pandora, your dog or your sanity as a result of this is solely
 *   YOUR responsibility.
 *
 *   You should have received a copy of the WTF Public License along with the
 *   source code to the game (please see the file COPYING).  If not, please refer to
 *   <http://www.wtfpl.net/about/>.
 */
#include "common.h"
#include "heroshot.h"
#include "enemy.h"
#include "explosion.h"
#include "blenders.h"
#include "hero.h"

#define  ENEMY_SIZE              7.5f
#define  ENEMY_OFFSCREEN_AMOUNT  32
#define  ENEMY_MAX_ENEMIES       256

#define  ENEMY_SPAWN_WEST        0
#define  ENEMY_SPAWN_NORTHWEST   1
#define  ENEMY_SPAWN_NORTH       2
#define  ENEMY_SPAWN_NORTHEAST   3
#define  ENEMY_SPAWN_EAST        4
#define  ENEMY_SPAWN_SOUTHEAST   5
#define  ENEMY_SPAWN_SOUTH       6
#define  ENEMY_SPAWN_SOUTHWEST   7

#define  ENEMY_NUM_DIRECTIONS    8

#define  ENEMY_LOGIC_COUNT_START 90
#define  ENEMY_LOGIC_COUNT_IDLE  45

#define  ENEMY_MOMENTUM_ATTENUATION  0.9875f


#define  ENEMY_TYPE1_SPEED       1.35f
#define  ENEMY_TYPE2_SPEED       0.0875f
#define  ENEMY_TYPE3_SPEED       1.45f

#define  ENEMY_MAX_TYPES         3
#define  ENEMY_INACTIVE          9999

/******************************************************************************/

typedef struct
{
   unsigned type;
   unsigned logic_counter;
   float    x;
   float    y;
   float    x_speed;
   float    y_speed;
   float    direction;
}
ENEMY;

static ENEMY enemy_pool[ENEMY_MAX_ENEMIES];
static int   enemy_pool_index = 0;
static int   enemy_from_which_dir = 0;

/******************************************************************************/

static BOOL enemy_art_loaded_yet = FALSE;
static BITMAP *enemy_gfx1 = NULL;
static BITMAP *enemy_gfx2 = NULL;
static SAMPLE *enemy_spawn_snd;

/******************************************************************************/
/*!
 * @brief Load all assets required to display enemies.
 */
void ENEMY_load(void)
{
   char buf[FILENAME_BUFFER_LENGTH];

   snprintf(buf, FILENAME_BUFFER_LENGTH, "./gfx/enemy/0.pcx");

   if (exists(buf))
   {
      enemy_gfx1 = load_bitmap(buf, NULL);
   }
   else
   {
      enemy_gfx1 = create_bitmap(16, 16);
      clear_to_color(enemy_gfx1, rand());
      DUH_WHERE_AM_I("missing asset: %s", buf);
   }

   snprintf(buf, FILENAME_BUFFER_LENGTH, "./gfx/enemy/1.pcx");

   if (exists(buf))
   {
      enemy_gfx2 = load_bitmap(buf, NULL);
   }
   else
   {
      enemy_gfx2 = create_bitmap(16, 16);
      clear_to_color(enemy_gfx2, rand());
      DUH_WHERE_AM_I("missing asset: %s", buf);
   }

   snprintf(buf, FILENAME_BUFFER_LENGTH, "./snd/enemy/spawn.wav");

   if (exists(buf))
   {
      enemy_spawn_snd = load_sample(buf);
   }
   else
   {
      enemy_spawn_snd = create_sample(8, 1, 22050, 1);
      DUH_WHERE_AM_I("missing asset: %s", buf);
   }
}

/******************************************************************************/
/*!
 * @brief Unload any assets used to display the enemies and free up memory.
 */
void ENEMY_unload(void)
{
   destroy_bitmap(enemy_gfx1);
   destroy_bitmap(enemy_gfx2);
   destroy_sample(enemy_spawn_snd);

   enemy_gfx1 = NULL;
   enemy_gfx2 = NULL;
}

/******************************************************************************/
/*!
 * @brief Decides off which side of the display to spawn the next enemy, as
 * well as the next enemy's type.
 * @note Not designed to be called from client code and not exposed in the
 * header.
 */
static void ENEMY_spawn_helper(int from_which_dir, int type, int index)
{
   switch (from_which_dir)
   {
      case ENEMY_SPAWN_WEST:
         enemy_pool[index].x = -ENEMY_OFFSCREEN_AMOUNT - ENEMY_SIZE;
         enemy_pool[index].y = rand() % DISPLAY_HEIGHT;
      break;

      case ENEMY_SPAWN_NORTHWEST:
         enemy_pool[index].x = -(rand() % ENEMY_OFFSCREEN_AMOUNT) - ENEMY_SIZE;
         enemy_pool[index].y = -(rand() % ENEMY_OFFSCREEN_AMOUNT) - ENEMY_SIZE;
      break;

      case ENEMY_SPAWN_NORTH:
         enemy_pool[index].x = rand() % DISPLAY_WIDTH;
         enemy_pool[index].y = -ENEMY_OFFSCREEN_AMOUNT - ENEMY_SIZE;
      break;

      case ENEMY_SPAWN_NORTHEAST:
         enemy_pool[index].x =  (rand() % ENEMY_OFFSCREEN_AMOUNT) + ENEMY_SIZE + DISPLAY_WIDTH;
         enemy_pool[index].y = -(rand() % ENEMY_OFFSCREEN_AMOUNT) - ENEMY_SIZE;
      break;

      case ENEMY_SPAWN_EAST:
         enemy_pool[index].x = ENEMY_OFFSCREEN_AMOUNT + ENEMY_SIZE;
         enemy_pool[index].y = rand() % DISPLAY_HEIGHT;
      break;

      case ENEMY_SPAWN_SOUTHEAST:
         enemy_pool[index].x =  (rand() % ENEMY_OFFSCREEN_AMOUNT) + ENEMY_SIZE + DISPLAY_WIDTH;
         enemy_pool[index].y =  (rand() % ENEMY_OFFSCREEN_AMOUNT) + ENEMY_SIZE + DISPLAY_HEIGHT;
      break;

      case ENEMY_SPAWN_SOUTH:
         enemy_pool[index].x = rand() % DISPLAY_WIDTH;
         enemy_pool[index].y = ENEMY_OFFSCREEN_AMOUNT + ENEMY_SIZE;
      break;

      case ENEMY_SPAWN_SOUTHWEST:
         enemy_pool[index].x = -(rand() % ENEMY_OFFSCREEN_AMOUNT) - ENEMY_SIZE;
         enemy_pool[index].y =  (rand() % ENEMY_OFFSCREEN_AMOUNT) + ENEMY_SIZE + DISPLAY_HEIGHT;
      break;
   }

   enemy_pool[index].logic_counter = ENEMY_LOGIC_COUNT_START;
   enemy_pool[index].type = type;
}

/******************************************************************************/
/*!
 * @brief Spawns a new enemy. The initial position, initial speed, and type
 * are decided automatically.
 */
void ENEMY_spawn(void)
{
   if (enemy_pool_index == ENEMY_MAX_ENEMIES) return;

   enemy_from_which_dir++;
   enemy_from_which_dir = enemy_from_which_dir % ENEMY_NUM_DIRECTIONS;

   ENEMY_spawn_helper(enemy_from_which_dir, rand() % ENEMY_MAX_TYPES, enemy_pool_index);

   if (enemy_spawn_snd != NULL)
   {
      float pan_value = (enemy_pool[enemy_pool_index].x - (DISPLAY_WIDTH / 2.0f)) + 128.0f;

      stop_sample(enemy_spawn_snd);
      play_sample(enemy_spawn_snd, 255, (int)pan_value , 1000, FALSE);
   }

   enemy_pool_index++;
}

/******************************************************************************/
/*!
 * @brief Deactivate the enemy with the specified index.
 * @note A new one will be spawned shortly, as game rules require there to
 * be a set number of enemies active to keep the pacing and difficulty right.
 */
void ENEMY_kill(int victim_index)
{
   EXPLOSION_spawn(enemy_pool[victim_index].x, enemy_pool[victim_index].y);

   enemy_from_which_dir++;
   enemy_from_which_dir = enemy_from_which_dir % ENEMY_NUM_DIRECTIONS;

   ENEMY_spawn_helper(enemy_from_which_dir, rand() % ENEMY_MAX_TYPES, victim_index);
}

/******************************************************************************/
/*!
 * @brief Deactivate ALL the active enemies; this will usually only happen after
 * the player has touched the warp or if we're ending play.
 */
void ENEMY_reset_all(void)
{
   int index;

   for (index = 0; index < ENEMY_MAX_ENEMIES; index++)
   {
      enemy_pool[index].type = ENEMY_INACTIVE;
      enemy_pool[index].x    = -999;
      enemy_pool[index].y    = -999;
   }

   enemy_pool_index = 0;
}

/******************************************************************************/
/*!
 * @brief Cause all enemies to advance their logic.
 * @note Must be called every frame.
 */
BOOL ENEMY_tick(void)
{
   BOOL was_player_hit = FALSE;
   int index = 0;

   float hero_x;
   float hero_y;

   HERO_get_pos(&hero_x, &hero_y);

   for (index = 0; index < enemy_pool_index; index++)
   {
      // were we hit by the player's gun?
      if (HEROSHOT_check_collision(enemy_pool[index].x, enemy_pool[index].y, ENEMY_SIZE))
      {
         ENEMY_kill(index);
      }

      // did we collide with the player?
      if (hypot(hero_x - enemy_pool[index].x, hero_y - enemy_pool[index].y) <=
            (HERO_INNER_RADIUS + ENEMY_SIZE))
      {
         ENEMY_kill(index);
         was_player_hit = TRUE;
      }

      float direction_to_hero = atan2(enemy_pool[index].y - hero_y,
         enemy_pool[index].x - hero_x);

      switch (enemy_pool[index].type)
      {
         case 0:
            // always chase, slow
            enemy_pool[index].x_speed = -cos(direction_to_hero) * ENEMY_TYPE1_SPEED;
            enemy_pool[index].y_speed = -sin(direction_to_hero) * ENEMY_TYPE1_SPEED;
         break;

         case 2:
            // always chase, fast
            enemy_pool[index].x_speed = -cos(direction_to_hero) * ENEMY_TYPE3_SPEED;
            enemy_pool[index].y_speed = -sin(direction_to_hero) * ENEMY_TYPE3_SPEED;
         break;

         case 1:
            // look at hero, then sprint towards where they last saw them, fast
            if (enemy_pool[index].logic_counter == ENEMY_LOGIC_COUNT_START)
            {
               enemy_pool[index].direction = direction_to_hero;
            }
            else if (enemy_pool[index].logic_counter > ENEMY_LOGIC_COUNT_IDLE)
            {
               enemy_pool[index].x_speed += -cos(enemy_pool[index].direction) * ENEMY_TYPE2_SPEED;
               enemy_pool[index].y_speed += -sin(enemy_pool[index].direction) * ENEMY_TYPE2_SPEED;
            }

            enemy_pool[index].logic_counter--;

            if (enemy_pool[index].logic_counter == 0)
            {
               enemy_pool[index].logic_counter = ENEMY_LOGIC_COUNT_START;
            }

            enemy_pool[index].x_speed *= ENEMY_MOMENTUM_ATTENUATION;
            enemy_pool[index].y_speed *= ENEMY_MOMENTUM_ATTENUATION;
         break;
      }

      enemy_pool[index].x += enemy_pool[index].x_speed;
      enemy_pool[index].y += enemy_pool[index].y_speed;
   }

   return was_player_hit;
}

/******************************************************************************/
/*!
 * @brief Repaint all active enemies.
 * @note Must be called every frame.
 */
void ENEMY_draw(void)
{
   int index;

   set_blender_mode(NULL, (BLENDER_FUNC)BLEND_add16, NULL, 0, 0, 0, 64);

   for (index = 0; index < enemy_pool_index; index++)
   {
      switch (enemy_pool[index].type)
      {
         case 0:
            draw_trans_sprite(double_buffer, enemy_gfx1,
               enemy_pool[index].x - (enemy_gfx1->w / 2),
               enemy_pool[index].y - (enemy_gfx1->h / 2));
         break;

         case 1:
            draw_trans_sprite(double_buffer, enemy_gfx2,
               enemy_pool[index].x - (enemy_gfx2->w / 2),
               enemy_pool[index].y - (enemy_gfx2->h / 2));
         break;

         case 2:
            draw_trans_sprite(double_buffer, enemy_gfx1,
               enemy_pool[index].x - (enemy_gfx1->w / 2),
               enemy_pool[index].y - (enemy_gfx1->h / 2));
         break;
      }
   }
}
