/******************************************************************************
/*   This file is part of the game Various Burst.
 *
 *   Various Burst is free software: you can redistribute it and/or modify it
 *   under the terms of the WTF Public License.
 *
 *   Various Burst is distributed in the hope that players will find it enjoyable,
 *   emotive, educational even, but with NO WARRANTY of any kind. What happens
 *   to your Pandora, your dog or your sanity as a result of this is solely
 *   YOUR responsibility.
 *
 *   You should have received a copy of the WTF Public License along with the
 *   source code to the game (please see the file COPYING).  If not, please refer to
 *   <http://www.wtfpl.net/about/>.
 */
#include "common.h"
#include "blenders.h"
#include "explosion.h"

#define  EXPLOSION_NUM_FRAMES          7
#define  EXPLOSION_NUM_EXPLOSIONS      32

typedef struct
{
   int x;
   int y;
   int frame;
}
EXPLOSION;

/*! @brief An array to track the explosion sprites.
 *  @note Originally, these were in a list, but there are so few of them, and
 *  they're so small, it seemed pointless...
 */
EXPLOSION   explosion_pool[EXPLOSION_NUM_EXPLOSIONS];

/*! @brief The actual explosion images. */
static BITMAP *explosion_gfx[EXPLOSION_NUM_FRAMES];

/*! @brief The explosion noise. */
static SAMPLE *explosion_sound;

/******************************************************************************/
/*!
 * @brief Load all the assets needed to display the enemy explosion effect.
 */
void EXPLOSION_load(void)
{
   char buf[FILENAME_BUFFER_LENGTH];
   int index;

   for (index = 0; index < EXPLOSION_NUM_FRAMES; index++)
   {
      snprintf(buf, FILENAME_BUFFER_LENGTH, "./gfx/explosion/%02x.pcx", index);

      if (exists(buf))
      {
         explosion_gfx[index] = load_bitmap(buf, NULL);
      }
      else
      {
         explosion_gfx[index] = create_bitmap(1, 1);
         DUH_WHERE_AM_I("missing asset: %s", buf);
      }
   }

   snprintf(buf, FILENAME_BUFFER_LENGTH, "./snd/explosion/0.wav");

   if (exists(buf))
   {
      explosion_sound = load_sample(buf);
   }
   else
   {
      explosion_sound = create_sample(8, 1, 22050, 1);
      DUH_WHERE_AM_I("missing asset: %s", buf);
   }
}

/******************************************************************************/
/*!
 * @brief Free up any resources used by the enemy explosion effect.
 */
void EXPLOSION_unload(void)
{
   int index;

   for (index = 0; index < EXPLOSION_NUM_FRAMES; index++)
   {
      destroy_bitmap(explosion_gfx[index]);
   }

   destroy_sample(explosion_sound);
}

/******************************************************************************/
/*!
 * @brief Halt any ongoing explosions; this usually only needs to happen when
 * ending play.
 */
void EXPLOSION_reset_all(void)
{
   int index;

   for (index = 0; index < EXPLOSION_NUM_EXPLOSIONS; index++)
   {
      explosion_pool[index].frame = EXPLOSION_NUM_FRAMES;
   }
}

/******************************************************************************/
/*!
 * @brief Spawn an explosion at the specified coordinates.  The explosion
 * animations will manage themselves to completion.
 * @param x The horizontal position the explosion should be started at.
 * @param y The vertical position the explosion should be started at.
 */
void EXPLOSION_spawn(int x, int y)
{
   static int prev_index;

   explosion_pool[prev_index].x = x;
   explosion_pool[prev_index].y = y;
   explosion_pool[prev_index].frame = 0;

   if (explosion_sound != NULL)
   {
      float pan_value = (explosion_pool[prev_index].x - (DISPLAY_WIDTH / 2.0f)) + 128.0f;

      stop_sample(explosion_sound);
      play_sample(explosion_sound, 255, (int)pan_value , 1000, FALSE);
   }

   prev_index++;
   prev_index = prev_index % EXPLOSION_NUM_EXPLOSIONS;
}

/******************************************************************************/
/*!
 * @brief Cause all ongoing explosions to advance their logic.
 * @note Must be called every frame.
 */
void EXPLOSION_tick(void)
{
   int index;

   for (index = 0; index < EXPLOSION_NUM_EXPLOSIONS; index++)
   {
      if (explosion_pool[index].frame < EXPLOSION_NUM_FRAMES)
      {
         explosion_pool[index].frame++;
      }
   }
}

/******************************************************************************/
/*!
 * @brief Cause all ongoing explosions to be painted.
 * @note Must be called every frame.
 */
void EXPLOSION_draw(void)
{
   int index;

   set_blender_mode(NULL, (BLENDER_FUNC)BLEND_add16, NULL, 0, 0, 0, 64);

   for (index = 0; index < EXPLOSION_NUM_EXPLOSIONS; index++)
   {
      if (explosion_pool[index].frame < EXPLOSION_NUM_FRAMES)
      {
         draw_trans_sprite(double_buffer, explosion_gfx[explosion_pool[index].frame],
         explosion_pool[index].x - (explosion_gfx[explosion_pool[index].frame]->w / 2),
         explosion_pool[index].y - (explosion_gfx[explosion_pool[index].frame]->h / 2));
      }
   }
}
