/******************************************************************************
/*   This file is part of the game Various Burst.
 *
 *   Various Burst is free software: you can redistribute it and/or modify
 *   it under the terms of the WTF Public License.
 *
 *   Various Burst is distributed in the hope that players will find it
 *   enjoyable, emotive, educational even, but with NO WARRANTY of any
 *   kind.  What happens to your computer, your dog or your sanity as
 *   a result of this is solely YOUR responsibility.
 *
 *   You should have received a copy of the WTF Public License
 *   along with the game (please see the file COPYING).  If not,
 *   please refer to <http://www.wtfpl.net/about/>.
 */

#include "common.h"
#include "input.h"
#include "hero.h"
#include "backdrop.h"
#include "warpeffect.h"
#include "hurteffect.h"
#include "pickups.h"
#include "gameplay.h"
#include "enemy.h"
#include "heroshot.h"
#include "deadeffect.h"
#include "highscore.h"
#include "nameentry.h"
#include "burst.h"

#define  GAMEPLAY_REQD_COINS_START  4
#define  GAMEPLAY_REQD_COINS_ADD    2

#define  GAMEPLAY_HEALTH_START      100
#define  GAMEPLAY_HEALTH_ENEMY_LOSS 8
#define  GAMEPLAY_HEALTH_COIN_ADD   10
#define  GAMEPLAY_HEALTH_LVL_ADD    25

#define  GAMEPLAY_COIN_TIMER_SPAWN  75
#define  GAMEPLAY_COIN_TIMER_END    195

static int gameplay_coins_required;
static int gameplay_coins_got;
static int gameplay_curr_lvl;
static int gameplay_curr_health;

static int gameplay_num_enemies_allowed;
static int gameplay_num_enemies_have;

static int gameplay_coin_timer;
static int gameplay_healthloss_timer;

static int  gameplay_warp_voice_num = -1;
static BOOL gameplay_anim_warp_up = FALSE;
static BOOL gameplay_level_up_anim_active = FALSE;
static BOOL gameplay_name_entry_active    = FALSE;

static SAMPLE *gameplay_warp_up_sound     = NULL;
static SAMPLE *gameplay_warp_down_sound   = NULL;

static int gameplay_timecount_clock;
static int gameplay_total_time_survived;
static int gameplay_total_stages_completed;
static int gameplay_total_cores_gotten;

/******************************************************************************/

static void GAMEPLAY_got_coin_handler(void)
{
   gameplay_coins_got++;
   gameplay_curr_health += GAMEPLAY_HEALTH_COIN_ADD;

   if (gameplay_coins_got % 2) gameplay_num_enemies_allowed++;

   if (gameplay_coins_got >= gameplay_coins_required)
   {
      PICKUPS_spawn_exit();
   }

   // scoring
   gameplay_total_cores_gotten++;
}

/******************************************************************************/

static void GAMEPLAY_got_exit_handler(void)
{
   gameplay_curr_lvl++;
   gameplay_coins_got = 0;
   gameplay_coins_required += GAMEPLAY_REQD_COINS_ADD;
   PICKUPS_remove();

   // start the warp sound and anim
   gameplay_level_up_anim_active = TRUE;
   gameplay_warp_voice_num = play_sample(gameplay_warp_up_sound, 255, 128, 1000, 0);
   WARPEFFECT_set(TRUE);

   gameplay_healthloss_timer = 0;
   gameplay_curr_health += GAMEPLAY_HEALTH_LVL_ADD;
   gameplay_anim_warp_up = TRUE;
   gameplay_num_enemies_have    = 0;

   gameplay_num_enemies_allowed = 1 + (gameplay_curr_lvl * 2);

   // scoring
   gameplay_total_stages_completed++;
}

/******************************************************************************/

void GAMEPLAY_load(void)
{
   ENEMY_load();
   EXPLOSION_load();
   HERO_load();
   PICKUPS_load();
   WARPEFFECT_load();
   BACKDROP_load_new(1);
   DEADEFFECT_load();
   NAME_load();
   BURST_load();

   char buf[FILENAME_BUFFER_LENGTH];

   snprintf(buf, FILENAME_BUFFER_LENGTH, "./snd/misc/warp-up.wav");

   if (exists(buf))
   {
      gameplay_warp_up_sound = load_sample(buf);
   }
   else
   {
      gameplay_warp_up_sound = create_sample(8, 0, 22050, 1);
      DUH_WHERE_AM_I("missing asset: %s", buf);
   }

   snprintf(buf, FILENAME_BUFFER_LENGTH, "./snd/misc/warp-down.wav");

   if (exists(buf))
   {
      gameplay_warp_down_sound = load_sample(buf);
   }
   else
   {
      gameplay_warp_down_sound = create_sample(8, 0, 22050, 1);
      DUH_WHERE_AM_I("missing asset: %s", buf);
   }

   // set the pickup handlers
   PICKUPS_set_coin_hit_callback(GAMEPLAY_got_coin_handler);
   PICKUPS_set_exit_hit_callback(GAMEPLAY_got_exit_handler);
}

/******************************************************************************/

void GAMEPLAY_unload(void)
{
   DEADEFFECT_unload();
   ENEMY_unload();
   EXPLOSION_unload();
   HERO_unload();
   PICKUPS_unload();
   WARPEFFECT_unload();
   BACKDROP_unload();
   NAME_unload();
   HURTEFFECT_reset();
   BURST_unload();

   destroy_sample(gameplay_warp_up_sound);
   destroy_sample(gameplay_warp_down_sound);
}

/******************************************************************************/

void GAMEPLAY_init(void)
{
   gameplay_name_entry_active = FALSE;

   gameplay_curr_health = GAMEPLAY_HEALTH_START;
   gameplay_curr_lvl    = 1;

   gameplay_num_enemies_allowed = 3;
   gameplay_num_enemies_have    = 0;

   gameplay_coins_required = GAMEPLAY_REQD_COINS_START;
   gameplay_coins_got      = 0;
   gameplay_coin_timer     = 0;

   gameplay_healthloss_timer = 0;

   gameplay_level_up_anim_active = FALSE;
   gameplay_anim_warp_up = FALSE;
   gameplay_warp_voice_num = -1;

   HERO_init();
   ENEMY_reset_all();
   WARPEFFECT_set(FALSE);
   EXPLOSION_reset_all();
   DEADEFFECT_init();
   NAME_init();
   BURST_reset_all();

   // scoring
   gameplay_total_time_survived     = 0;
   gameplay_total_stages_completed  = 0;
   gameplay_total_cores_gotten      = 0;

   // clock to mark tenths of a second
   gameplay_timecount_clock = 0;

   COMMON_set_bgm(1);

   play_sample(gameplay_warp_down_sound, 255, 128, 1000, 0);
}

/******************************************************************************/

void GAMEPLAY_tick(void)
{
   float tmp_x;
   float tmp_y;

   int hs_tmp;

   DEADEFFECT_tick();
   BURST_tick();

   if (gameplay_curr_health > GAMEPLAY_HEALTH_START)
   {
      gameplay_curr_health = GAMEPLAY_HEALTH_START;
   }

   WARPEFFECT_grab();

   EXPLOSION_tick();

   if (!gameplay_level_up_anim_active)
   {
      // normal gameplay ------------------------------------------------
      HEROSHOT_tick();

      // only allow stuff to happen if the hero isn't dead...
      if (gameplay_curr_health >= 0)
      {
         // manage coins
         if (gameplay_coins_got != gameplay_coins_required)
         {
            gameplay_coin_timer++;

            if (gameplay_coin_timer == GAMEPLAY_COIN_TIMER_SPAWN)
            {
               PICKUPS_spawn_coin();
            }

            if (gameplay_coin_timer >= GAMEPLAY_COIN_TIMER_END)
            {
               PICKUPS_remove();
               gameplay_coin_timer = 0;
            }
         }

         // manage hero
         HERO_tick(JOY_UP, JOY_DOWN, JOY_LEFT, JOY_RIGHT, ((FIRE_B == BTN_PRESSED) || (FIRE_X == BTN_PRESSED)), 0);

         HERO_get_pos(&tmp_x, &tmp_y);

         PICKUPS_tick(tmp_x, tmp_y, HERO_OUTER_RADIUS);
         if (ENEMY_tick())
         {
            if (HURTEFFECT_check_done())
            {
               HURTEFFECT_start();
               gameplay_curr_health -= GAMEPLAY_HEALTH_ENEMY_LOSS;
            }
         }

         // keep time
         gameplay_timecount_clock++;

         if (gameplay_timecount_clock == 6)
         {
            gameplay_timecount_clock = 0;
            gameplay_total_time_survived++;
         }

         gameplay_healthloss_timer++;

         if (gameplay_healthloss_timer == TICKS_PER_SEC)
         {
            gameplay_healthloss_timer = 0;
            gameplay_curr_health--;

            if (gameplay_num_enemies_allowed > gameplay_num_enemies_have)
            {
               gameplay_num_enemies_have++;
               ENEMY_spawn();
            }
         }
      }
   }
   else
   {
      if (voice_check(gameplay_warp_voice_num) == NULL)
      {
         if (gameplay_anim_warp_up)
         {
            // warp up is done, warp down now
            gameplay_anim_warp_up = FALSE;
            gameplay_warp_voice_num = play_sample(gameplay_warp_down_sound, 255, 128, 1000, 0);
            WARPEFFECT_set(FALSE);
            BACKDROP_load_new(gameplay_curr_lvl);
            ENEMY_reset_all();
         }
         else
         {
            HERO_stop();
            gameplay_level_up_anim_active = FALSE;
         }
      }
   }

   // stat screen
   // ...admittedly, there was a better way to do this.
   if (gameplay_curr_health < 0)
   {
      COMMON_fade_bgm();
      DEADEFFECT_start();

      if (DEADEFFECT_get_state() == DEADEFFECT_STATE_DONE)
      {
         hs_tmp = HIGHSCORE_compute_total(gameplay_total_stages_completed, gameplay_total_time_survived, gameplay_total_cores_gotten);

         if (HIGHSCORE_check_ranked(hs_tmp))
         {
            // earned a spot on the list
            NAME_start();
         }
         else
         {
            // didn't
            curr_state_done   = TRUE;
            next_state        = STATE_MENU;
         }

         // manage the name entry widget
         if (NAME_check_done())
         {
            HIGHSCORE_append(hs_tmp, NAME_fetch());
            curr_state_done   = TRUE;
            next_state        = STATE_MENU;
         }
         else
         {
            NAME_tick();
         }
      }
   }
}

/******************************************************************************/

void GAMEPLAY_draw(void)
{
   static int blink;

   blink++;

   BACKDROP_draw();

   if (gameplay_curr_health > 0)
   {
      HERO_draw();
      ENEMY_draw();
      PICKUPS_draw();
      HEROSHOT_draw();
      EXPLOSION_draw();
      WARPEFFECT_draw();
      HURTEFFECT_draw();
      BURST_draw();

      textprintf_ex(double_buffer, font, 200, 16, -1, -1,  "STAGE $%02X", gameplay_curr_lvl);
      textprintf_ex(double_buffer, font, 16, 16, -1, -1,  "HEALTH:");
      rectfill(double_buffer, 97, 16, 196, 24, makecol(255, 0, 0));
      rectfill(double_buffer, 97, 16, 96 + gameplay_curr_health, 24, makecol(32, 255, 128));


      if (!WARPEFFECT_check_enabled())
      {
         if (gameplay_coins_required != gameplay_coins_got)
         {
            textprintf_ex(double_buffer, font, 16, 224, -1, -1, "collect %02d cores to advance", gameplay_coins_required - gameplay_coins_got);
         }
         else
         {
            textprintf_ex(double_buffer, font, 16, 224, -1, -1, "          PROCEED INTO THE WARP", gameplay_coins_required - gameplay_coins_got);
         }
      }
   }

   DEADEFFECT_draw();
   if (DEADEFFECT_get_state() == DEADEFFECT_STATE_WAIT_USER)
   {
      textprintf_centre_ex(double_buffer, font, DISPLAY_WIDTH / 2, 176, -1, -1,  "survived: %02.1f SEC", gameplay_total_time_survived / 10.0f);
      textprintf_centre_ex(double_buffer, font, DISPLAY_WIDTH / 2, 184, -1, -1,  "cores colleced: %02d", gameplay_total_cores_gotten);
      textprintf_centre_ex(double_buffer, font, DISPLAY_WIDTH / 2, 192, -1, -1,  "stages completed: %02d", gameplay_total_stages_completed);
      textprintf_centre_ex(double_buffer, font, DISPLAY_WIDTH / 2, 200, -1, -1,  "YOUR SCORE: %08d",
         HIGHSCORE_compute_total(gameplay_total_stages_completed, gameplay_total_time_survived, gameplay_total_cores_gotten));


      if ((blink % 100) < 50) textprintf_centre_ex(double_buffer, font, DISPLAY_WIDTH / 2, 216, -1, -1,  "PRESS (B) TO ADVANCE");
   }

   if (DEADEFFECT_get_state() == DEADEFFECT_STATE_DONE)
   {
      NAME_draw();
   }
}
