/******************************************************************************/
/*   This file is part of the game Various Burst.
 *
 *   Various Burst is free software: you can redistribute it and/or modify
 *   it under the terms of the WTF Public License.
 *
 *   Various Burst is distributed in the hope that players will find it
 *   enjoyable, emotive, educational even, but with NO WARRANTY of any
 *   kind.  What happens to your computer, your dog or your sanity as
 *   a result of this is solely YOUR responsibility.
 *
 *   You should have received a copy of the WTF Public License
 *   along with the game (please see the file COPYING).  If not,
 *   please refer to <http://www.wtfpl.net/about/>.
 */

#include "common.h"
#include "hero.h"
#include "blenders.h"
#include "heroshot.h"

/*!
 * @defgroup hero_c The Hero
 * @brief Structures and functions for managing the hero (the player's ship).
 * @{
 */

/*! @brief How much to multiply the hero's speed by each tick to simulate drag
 *  or friction.
 */
#define  HERO_MOMENTUM_ATTENUATION  0.975f

/*! @brief The maximum speed on any coordinate plane the hero is allowed to
 * travel. (yes, it's possible to travel ~1.4 times the max speed diagonally -
 * this is as-designed...)
 */
#define  HERO_MAX_SPEED          2.75f

/*! @brief How much to accelerate the hero by when its told to move.
 */
#define  HERO_THRUST_ADD         .35f

/*! @brief How much the hero turns per frame when told to fly a different
 *  direction (in 15.16 allegrodegs).
 *  @warning This needs to be a factor of 256, or else animation glitches will
 *  occur.
 */
#define  HERO_TURN_AMOUNT        (8  << 16)

#define  HERO_DIR_NORTH          (0   << 16)
#define  HERO_DIR_NORTHEAST      (32  << 16)
#define  HERO_DIR_EAST           (64  << 16)
#define  HERO_DIR_SOUTHEAST      (96  << 16)
#define  HERO_DIR_SOUTH          (128 << 16)
#define  HERO_DIR_SOUTHWEST      (160 << 16)
#define  HERO_DIR_WEST           (192 << 16)
#define  HERO_DIR_NORTHWEST      (224 << 16)

/******************************************************************************/
/*! @brief A structure that describes a hero (only one of them is ever made). */
typedef struct
{
   float x;
   float y;
   float x_speed;
   float y_speed;
   int   direction;
   int   curr_heading;
   int   shot_timer;
   int   thrust_particle_timer;
   int   invincibility_timer;
   int   health;
   int   death_animation_timer;
} HERO_PRIV;

/*! @brief The actual instance of the hero. Should be considered module-private.
 */
static HERO_PRIV  hero;

/*! @brief The sprite images for the hero are stored in this array; should be
 * considered module-private.*/
static BITMAP    *hero_gfx;
static BITMAP    *hero_rot_buffer;

/******************************************************************************/
/*! @brief Load all the images for the hero.
 */
void HERO_load(void)
{
   char buf[FILENAME_BUFFER_LENGTH];
   snprintf(buf, FILENAME_BUFFER_LENGTH, "./gfx/hero/0.pcx");

   if (exists(buf))
   {
      hero_gfx = load_bitmap(buf, NULL);
   }
   else
   {
      hero_gfx = create_bitmap(16, 16);
      clear_to_color(hero_gfx, makecol(0, 255, 0));
      DUH_WHERE_AM_I("missing asset: %s", buf);
   }

   hero_rot_buffer = create_bitmap(36, 36);

   HEROSHOT_load();
}

/******************************************************************************/
/*! @brief Free up all the images for the hero.
 */
void HERO_unload(void)
{
   destroy_bitmap(hero_gfx);
   destroy_bitmap(hero_rot_buffer);
   HEROSHOT_unload();

}

/******************************************************************************/
/*! @brief Reset the hero to the centre of the display, along with resetting all
 *  logic counters, etc.
 */
void HERO_init(void)
{
   hero.x_speed = 0.0f;
   hero.y_speed = 0.0f;

   hero.x = DISPLAY_WIDTH / 2.0f;
   hero.y = DISPLAY_HEIGHT / 2.0f;

   hero.direction = HERO_DIR_NORTH;
}

/******************************************************************************/
/*! @brief Paint the hero with additive blending at its current location.
 */
void HERO_draw(void)
{
   clear_to_color(hero_rot_buffer, makecol(255, 0, 255));
   rotate_sprite(hero_rot_buffer, hero_gfx,
      (int)((hero_rot_buffer->w - hero_gfx->w) / 2.0f),
      (int)((hero_rot_buffer->h - hero_gfx->h) / 2.0f),
      hero.curr_heading);

   set_blender_mode(NULL, (BLENDER_FUNC)BLEND_add16, NULL, 0, 0, 0, 64);
   draw_trans_sprite(double_buffer, hero_rot_buffer,
      (int)(hero.x - (hero_rot_buffer->w / 2.0f)),
      (int)(hero.y - (hero_rot_buffer->h / 2.0f)));
}

/******************************************************************************/
/*! @brief Retrieves the current position of the hero in coordinate space.
 *  @param out_x The address of a float where the x-coordinate should be stored.
 *  @param out_y The address of a float where the y-coordinate should be stored.
 *  @note Either parameter may safely be NULL.
 */
void HERO_get_pos(float *out_x, float *out_y)
{
   if (out_x != NULL) *out_x = hero.x;
   if (out_y != NULL) *out_y = hero.y;
}

/******************************************************************************/
/*! @brief Advances the hero's movement, as well as advancing all logic counters.
 */
void HERO_tick(unsigned up, unsigned down, unsigned left, unsigned right, unsigned fire, unsigned evade)
{
   // directional control...
   if (up)
   {
      if (left)
      {
         hero.direction = HERO_DIR_NORTHWEST;
         hero.x_speed -= HERO_THRUST_ADD;
         hero.y_speed -= HERO_THRUST_ADD;
      }
      else if (right)
      {
         hero.direction = HERO_DIR_NORTHEAST;
         hero.x_speed += HERO_THRUST_ADD;
         hero.y_speed -= HERO_THRUST_ADD;
      }
      else
      {
         hero.direction = HERO_DIR_NORTH;
         hero.y_speed -= HERO_THRUST_ADD;
      }
   }
   else if (down)
   {
      if (left)
      {
         hero.direction = HERO_DIR_SOUTHWEST;
         hero.x_speed -= HERO_THRUST_ADD;
         hero.y_speed += HERO_THRUST_ADD;
      }
      else if (right)
      {
         hero.direction = HERO_DIR_SOUTHEAST;
         hero.x_speed += HERO_THRUST_ADD;
         hero.y_speed += HERO_THRUST_ADD;
      }
      else
      {
         hero.direction = HERO_DIR_SOUTH;
         hero.y_speed += HERO_THRUST_ADD;
      }
   }
   else if (left)
   {
      hero.direction = HERO_DIR_WEST;
      hero.x_speed -= HERO_THRUST_ADD;
   }
   else if (right)
   {
      hero.direction = HERO_DIR_EAST;
      hero.x_speed += HERO_THRUST_ADD;
   }

   // make the sprite swivel round smoothly over time...
   if (hero.curr_heading != hero.direction)
   {
      // convoluted logic here (sorry!). it's to make the
      // hero turn the least possible amount.
      if ((hero.direction == HERO_DIR_EAST) || (hero.direction == HERO_DIR_NORTHEAST) || (hero.direction == HERO_DIR_SOUTHEAST))
      {
         if (hero.curr_heading < hero.direction)
         {
            hero.curr_heading += HERO_TURN_AMOUNT;
         }
         else
         {
            if (hero.curr_heading < HERO_DIR_WEST)
            {
               hero.curr_heading -= HERO_TURN_AMOUNT;
            }
            else
            {
               hero.curr_heading += HERO_TURN_AMOUNT;
            }
         }
      }
      else
      if ((hero.direction == HERO_DIR_WEST) || (hero.direction == HERO_DIR_NORTHWEST))
      {
         if (hero.curr_heading > hero.direction)
         {
            hero.curr_heading -= HERO_TURN_AMOUNT;
         }
         else
         {
            if (hero.curr_heading < HERO_DIR_EAST)
            {
               hero.curr_heading -= HERO_TURN_AMOUNT;
            }
            else
            {
               hero.curr_heading += HERO_TURN_AMOUNT;
            }
         }
      }
      else if (hero.direction == HERO_DIR_SOUTHWEST)
      {
         // do NOT remove this special case - it's a bugfix.
         if (hero.curr_heading > hero.direction)
         {
            hero.curr_heading -= HERO_TURN_AMOUNT;
         }
         else
         {
            if (hero.curr_heading < HERO_DIR_NORTHEAST)
            {
               hero.curr_heading -= HERO_TURN_AMOUNT;
            }
            else
            {
               hero.curr_heading += HERO_TURN_AMOUNT;
            }
         }
      }
      else if (hero.direction == HERO_DIR_NORTH)
      {
         if (hero.curr_heading < HERO_DIR_SOUTH)
            hero.curr_heading -= HERO_TURN_AMOUNT;
         else
            hero.curr_heading += HERO_TURN_AMOUNT;
      }
      else
      {
         if (hero.curr_heading < HERO_DIR_SOUTH)
            hero.curr_heading += HERO_TURN_AMOUNT;
         else
            hero.curr_heading -= HERO_TURN_AMOUNT;
      }

      // catch the case where we're all the way around a cycle and into another
      if (hero.curr_heading == (256 << 16)) hero.curr_heading = 0;
      if (hero.curr_heading < 0           ) hero.curr_heading = (256 << 16) - HERO_TURN_AMOUNT;
   }

   if (hero.x_speed > HERO_MAX_SPEED) hero.x_speed = HERO_MAX_SPEED;
   if (hero.y_speed > HERO_MAX_SPEED) hero.y_speed = HERO_MAX_SPEED;

   if (hero.x_speed < -HERO_MAX_SPEED) hero.x_speed = -HERO_MAX_SPEED;
   if (hero.y_speed < -HERO_MAX_SPEED) hero.y_speed = -HERO_MAX_SPEED;

   hero.x += hero.x_speed;
   hero.y += hero.y_speed;

   if (hero.x < 0) hero.x = 0;
   if (hero.y < 0) hero.y = 0;
   if (hero.x > DISPLAY_WIDTH)   hero.x = DISPLAY_WIDTH;
   if (hero.y > DISPLAY_HEIGHT)  hero.y = DISPLAY_HEIGHT;

   hero.x_speed *= HERO_MOMENTUM_ATTENUATION;
   hero.y_speed *= HERO_MOMENTUM_ATTENUATION;

   if (fire)
   {
      HEROSHOT_fire(hero.x, hero.y, ALLEGRODEGS_TO_RADIANS(fixtoi(hero.curr_heading)));
   }
}

/******************************************************************************/

void HERO_stop(void)
{
   hero.x_speed = 0;
   hero.y_speed = 0;
}

/*! @} */
