/******************************************************************************
/*   This file is part of the game Various Burst.
 *
 *   Various Burst is free software: you can redistribute it and/or modify
 *   it under the terms of the WTF Public License.
 *
 *   Various Burst is distributed in the hope that players will find it
 *   enjoyable, emotive, educational even, but with NO WARRANTY of any
 *   kind.  What happens to your computer, your dog or your sanity as
 *   a result of this is solely YOUR responsibility.
 *
 *   You should have received a copy of the WTF Public License
 *   along with the game (please see the file COPYING).  If not,
 *   please refer to <http://www.wtfpl.net/about/>.
 */


#include "common.h"
#include "heroshot.h"
#define  HERO_SHOT_SPEED         7.5f
#define  HERO_SHOT_MAX_ACTIVE    16

/*! @defgroup heroshot_c Hero Shots
 * @brief Contains code to manage shots fired by the hero.
 * @{
 */

/*! @brief A structure that represents a shot fired by the hero.
 *  @note The 'next' pointer is vestigial and should remain commented out. The
 *  original implementation had the HERO_SHOTs in a list; because so few of them
 *  can physically exist at a time under the game rules, switched to a pool
 *  instead to reduce complexity.
 */
typedef struct
{
   BOOL  alive;
   int   col;
   float x;
   float y;
   float prev_x;
   float prev_y;
   float direction;
// void  *next;
}
HERO_SHOT;

static HERO_SHOT heroshot_array[HERO_SHOT_MAX_ACTIVE];
static SAMPLE *heroshot_sound = NULL;

/******************************************************************************/

void HEROSHOT_load(void)
{
   char buf[FILENAME_BUFFER_LENGTH];
   snprintf(buf, FILENAME_BUFFER_LENGTH, "./snd/hero/shot.wav");

   if (exists(buf))
   {
      heroshot_sound = load_sample(buf);
   }
   else
   {
      heroshot_sound = create_sample(8, 1, 22050, 1);
      DUH_WHERE_AM_I("missing asset: %s", buf);
   }
}

/******************************************************************************/

void HEROSHOT_unload(void)
{
   destroy_sample(heroshot_sound);
}


/******************************************************************************/
/*! @brief Fire a new shot.
 */
void HEROSHOT_fire(float start_x, float start_y, float direction)
{
   static int start_index;
   int strides = 0;

   while (strides < HERO_SHOT_MAX_ACTIVE)
   {
      if (!heroshot_array[(strides + start_index) % HERO_SHOT_MAX_ACTIVE].alive)
      {
         if (heroshot_sound != NULL)
         {
            float pan_value = (start_x - (DISPLAY_WIDTH / 2.0f)) + 128.0f;

            stop_sample(heroshot_sound);
            play_sample(heroshot_sound, 255, (int)pan_value , 1000, FALSE);
         }

         heroshot_array[(strides + start_index) % HERO_SHOT_MAX_ACTIVE].x        = start_x;
         heroshot_array[(strides + start_index) % HERO_SHOT_MAX_ACTIVE].y        = start_y;
         heroshot_array[(strides + start_index) % HERO_SHOT_MAX_ACTIVE].prev_x   = start_x;
         heroshot_array[(strides + start_index) % HERO_SHOT_MAX_ACTIVE].prev_y   = start_y;
         heroshot_array[(strides + start_index) % HERO_SHOT_MAX_ACTIVE].direction = direction;
         heroshot_array[(strides + start_index) % HERO_SHOT_MAX_ACTIVE].col      = 0;
         heroshot_array[(strides + start_index) % HERO_SHOT_MAX_ACTIVE].alive    = TRUE;
         start_index = (strides + start_index + 1) % HERO_SHOT_MAX_ACTIVE;
         return;
      }

      strides++;
   }
}

/******************************************************************************/

void HEROSHOT_tick(void)
{
   int index;

   for (index = 0; index < HERO_SHOT_MAX_ACTIVE; index++)
   {
      heroshot_array[index].col++;
      heroshot_array[index].prev_x = heroshot_array[index].x;
      heroshot_array[index].prev_y = heroshot_array[index].y;

      heroshot_array[index].x += sin(heroshot_array[index].direction) * HERO_SHOT_SPEED;
      heroshot_array[index].y += -cos(heroshot_array[index].direction) * HERO_SHOT_SPEED;

      // discard if outside the screen
      if ((heroshot_array[index].prev_x < 0) || (heroshot_array[index].prev_x > DISPLAY_WIDTH) ||
          (heroshot_array[index].prev_y < 0) || (heroshot_array[index].prev_y > DISPLAY_HEIGHT))
      {
         heroshot_array[index].alive = FALSE;
      }
   }
}

/******************************************************************************/

void HEROSHOT_draw(void)
{
   unsigned char colour_shift;
   int index;

   for (index = 0; index < HERO_SHOT_MAX_ACTIVE; index++)
   {
      colour_shift = (unsigned char)((cos(heroshot_array[index].col) + 1.0f) * 127.5f);
      line(double_buffer, heroshot_array[index].prev_x, heroshot_array[index].prev_y,
         heroshot_array[index].x, heroshot_array[index].y, makecol(colour_shift, colour_shift, 255));
   }
}

/******************************************************************************/

BOOL HEROSHOT_check_collision(float centre_x, float centre_y, float radius)
{
   float epsilon = 0.5f; // amount of fudge/inaccuracy we'll accept
   int index;

   for (index = 0; index < HERO_SHOT_MAX_ACTIVE; index++)
   {
      if (heroshot_array[index].alive)
      {
         // not sure if this is the best way to do this or not, but...
         float dist_from_point_0 = hypotf(heroshot_array[index].x - centre_x,
                                       heroshot_array[index].prev_y - centre_y);

         float dist_from_point_1 = hypotf(heroshot_array[index].x - centre_x,
                                       heroshot_array[index].prev_y - centre_y);

         if ((fabsf(dist_from_point_1 - dist_from_point_0) < epsilon) && // are the distances equal-ish?
             (dist_from_point_0 < HERO_SHOT_SPEED))                      // does the line sorta kind cross the sprite?
         {
            // scored a hit
            heroshot_array[index].alive = FALSE;
            return TRUE;
         }
      }
   }
   return FALSE;
}
